package server;

import java.util.*;
import java.io.*;
import java.net.*;

import data.*;

public class WorldServer {

    protected static String mapFilename = "default.map";
    protected static int serverPort = 9000;

    protected static String initialMoney  = "1000";
    protected static String robotCapacity = "100";

    private int width;
    private int height;
    private WorldMap worldMap;

    private Hashtable clients = new Hashtable();  // key is robotID, contains RobotClient

    private boolean gameStarted = false;

    private Hashtable commands = new Hashtable();

    public static void main(String[] args) {
        for (int i = 0; i < args.length; i++) {
            if (args[i].equalsIgnoreCase("-m")) {
                mapFilename = args[++i];
            } else
            if (args[i].equalsIgnoreCase("-p")) {
                serverPort = new Integer(args[++i]).intValue();
            }
        }
        try {
            new WorldServer();
        } catch (Exception bad) {
            bad.printStackTrace();
        }
    }


    public WorldServer() throws Exception {
        readWorldMap();
        new ClientThread().start();
    }

    private void readWorldMap() throws Exception {
        System.out.println("Loading map....");
        File mapFile = new File(mapFilename);
        BufferedReader reader = new BufferedReader(new FileReader(mapFile));
 
        String line = reader.readLine();
        StringTokenizer st = new StringTokenizer(line, " ");
        width = new Integer(st.nextToken()).intValue();
        height = new Integer(st.nextToken()).intValue();

        worldMap = new WorldMap(width, height);

        for (int row = height; row > 0; row--) {
            line = reader.readLine().trim();
            for (int col = 1; col <= line.length(); col++) {
                worldMap.addItem(col, row, line.charAt(col-1));
            }
        }

        // ID WEIGHT LOCX LOCY DESTX DESTY
        line = reader.readLine();
        while (line != null) {
            StringTokenizer item = new StringTokenizer(line, " ");
            String id = item.nextToken();
            int w = new Integer(item.nextToken()).intValue();
            int col = new Integer(item.nextToken()).intValue();
            int row = new Integer(item.nextToken()).intValue();
            int dCol = new Integer(item.nextToken()).intValue();
            int dRow = new Integer(item.nextToken()).intValue();
            worldMap.addPackage(new PackageItem(id, w, 0, 0, dCol, dRow), col, row);
            line = reader.readLine();
        }
        
        reader.close();
        
    }

    public void onClientMessage(String robotID, String message) {
        if (message.equals("Player")) {
            if (!gameStarted) {
                initializeClient(robotID);
            }
        } else {
            processRobotCommand(robotID, message);
        }
    }

    public void onClientError(String robotID, String message) {
    }

    private void initializeClient(String robotID) {
        RobotClient client = (RobotClient) clients.get(robotID);

        System.out.println("Initialize client...");

        client.sendMessage(width + " " + height);
        for (int row = 1; row <= height; row++) {
            client.sendMessage(worldMap.getRow(row, true));
        }

        client.sendMessage(robotID + " " + robotCapacity + " " + initialMoney);

        GridItem home = worldMap.findEmptyBase();
        if (home != null) {
            home.addRobot(Robot.getRobot(robotID));
        }
        client.sendMessage(worldMap.getRobotLocations());

        if (worldMap.allBasesOccupied()) {
            gameStarted = true;
            playGameTurn();
        } else System.out.println("NOT READY TO START");
    }

    private void processRobotCommand(String robotID, String command) {
        commands.put(robotID, command);

        // did we get a command from everyone?
        if (commands.size() == clients.size()) {
            runAllCommands();
        }
    }

    private void runAllCommands() {
        String turnResults = "";
        for (Enumeration c = commands.keys(); c.hasMoreElements(); ) {
            String robotID = (String) c.nextElement();
            String command = (String) commands.get(robotID);
            StringTokenizer st = new StringTokenizer(command, " ");

            int bid = new Integer(st.nextToken()).intValue();
            String action = st.nextToken();
            System.out.println("R " + robotID + " B " + bid + " A " + action);
            if (action.equals("Move")) {
                turnResults += worldMap.move(Robot.getRobot(robotID), st.nextToken());
            } else
            if (action.equals("Pick")) {
                String ids = "";
                while (st.hasMoreTokens()) {
                    ids += st.nextToken();
                    if (st.hasMoreTokens()) ids += " ";
                }
                turnResults += worldMap.pick(robotID, ids);
            } else
            if (action.equals("Drop")) {
                String ids = "";
                while (st.hasMoreTokens()) {
                    ids += st.nextToken();
                    if (st.hasMoreTokens()) ids += " ";
                }
                turnResults += worldMap.drop(robotID, ids);
            }
        }

        for (Enumeration c = clients.elements(); c.hasMoreElements(); ) {
            RobotClient client = (RobotClient) c.nextElement();
            client.sendMessage(turnResults);
        }

        System.out.println("TURN OVER: " + turnResults +"\n" + worldMap);

        if (worldMap.totalPackages() == 0) {
            System.out.println("ALL PACKAGES DELIVERED");
            System.exit(0);
        }
        
        // PLAY ANOTHER GAME TURN
        commands.clear();
        playGameTurn();
    }
    
    private void playGameTurn() {
        System.out.println("PLAY");
        // send the package information
        for (Enumeration c = clients.elements(); c.hasMoreElements(); ) {
            RobotClient client = (RobotClient) c.nextElement();
            client.sendMessage(worldMap.getPackageInfo(client.getID()));
        }
    }
    
    public class ClientThread extends Thread {

        public ClientThread() {
            super();
        }

        public void run() {
            try {
                ServerSocket server = new ServerSocket(serverPort);
                int robots = 0;
                
                System.out.println("Waiting for clients...");
                while (true) {
                    Socket socket = server.accept();
                    robots++;

                    String robotID = new Integer(robots).toString();
                    System.out.println("Contacted by " + robotID);
                    RobotClient client = new RobotClient(WorldServer.this, robotID, socket);
                    clients.put(robotID, client);
                }
            } catch (Exception ignore) {
                ignore.printStackTrace();
            }
        }
    }
}
